/* An external clock sync utility for IM style mailers,
	Michiel van der Vlist (2:500/9.5) */

#include <io.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <ctype.h>
#include <fcntl.h>
#include <values.h>
#include <dos.h>

#define TRUE (1==1)
#define FALSE (!TRUE)
#define PROGNAME "MTsync"
#define VERSION "1.00"

#define MYID "(c) 2000, Michiel van der Vlist, PA0MMV (2:280/5555)"
#define INLOGFILE "IM.LOG"
#define OUTLOGFILE "MTSYNC.LOG"
#define LOGBUFSIZ 16000
#define NRHISAKAS 80
#define NRMATCHAKAS 20
#define MAXAGE 7200
#define MAXCONNECT 3
#define MAXDIFF 900

typedef char string[128];
typedef char boolean;
typedef struct { unsigned int zone,net,node,point;} nodenumber;

void help(void);
int getcommands(int nargs, char *args[]);
boolean seek_connect_back(int *pos);
boolean seekdate_back(int *pos, boolean *ring);
boolean seek_emsi(int *pos, int buflen, boolean ring);
long int seek_tranx(int *trxpos,int buflen);
long int get_unix_time(int datepos, int emsipos);
long int seek_tzutc(int *tzpos,int buflen);
void pstring(char *st,int n);
void logheader(void);
void timestamp(FILE *f, int s);
int scanfidonumber(char *s, nodenumber *f);
void gethisakas(int emsipos);
int akasmatch(void);

nodenumber hisakas[NRHISAKAS];
nodenumber matchaka[NRMATCHAKAS];
nodenumber Mynode, Lastnode;
int nrhisakas, nrmatchakas=0;

const char *monthn[]={"***","Jan","Feb","Mar","Apr","May","Jun",
									"Jul","Aug","Sep","Oct","Nov","Dec"};
const char *dow[]={"Sun","Mon","Thu","Wed","Thu","Fri","Sat"};

string pstr,Inputlog,Outputlog,Semaphore;
boolean logflag=1, logisopen=FALSE, Rtcflag=FALSE;
char logbuf[LOGBUFSIZ];
int Maxconnects=MAXCONNECT;
long int Maxage=MAXAGE, Maxdiff=MAXDIFF;
FILE *Outlog;

int main(int argc, char *argv[])
{
	char *mytzutc, *envtz;
	string envtzsave="TZ=";
	long int t,flength, startpos, trx, timediff, td,tzutc=MAXLONG,mytz=0;
	long int sematime=0;
	time_t timer;
	struct time d_time;
	struct date d_date;
	struct ftime ftimep;
	int i,j,logfile,sema,connectpos,buflen,connectsfound=0;
	int emsipos,datepos,tranxpos,tzutcpos;
	boolean ring;
	FILE *fsema;
	union REGS regs;


	Lastnode.zone=2;
	Lastnode.net=Lastnode.node=Lastnode.point=0;

	printf("\nMTsync Clock synchroniser %s\n",VERSION);
	printf("%s\n",MYID);

	strcpy(Inputlog,INLOGFILE);
	strcpy(Outputlog,OUTLOGFILE);
	strcpy(Semaphore,PROGNAME); strcat(Semaphore,".NOW");

	/* Some wizzardry is needed to program around Borland's time zone
		shit. Apparently the dostounix function assumes we are at EST
		if no (valid) TZ variable is defined. This kludge appears to do
		the trick... */

	if ((envtz=getenv("TZ"))!=NULL) strncat(envtzsave,envtz,sizeof(envtz)-4);
	if (putenv("TZ=UTC0")==-1)
	{
		sprintf(pstr,"No room in environment to set TZ variable.\n");
		pstring(pstr,1);
	}
	tzset();

	if (!getcommands(argc,argv)) return(0);

	printf("Opening %s for read.\n",Inputlog);
	logfile=open(Inputlog,O_TEXT|O_RDONLY|O_DENYNONE);
	if (logfile==-1)
	{
		printf("Can't open %s for read. %s Aborted!\n",Inputlog,PROGNAME);
		return(2);
	}
	flength=filelength(logfile);
	if (flength>LOGBUFSIZ) buflen=LOGBUFSIZ; else buflen=(int)flength;
	startpos=flength-buflen;
	lseek(logfile,startpos,SEEK_SET);
	read(logfile,logbuf,buflen);
	close(logfile);

	logheader();
	sprintf(pstr,"%s succesfully read.\n",Inputlog);
	pstring(pstr,0);
	sema=open(Semaphore,O_RDONLY);
	if (sema!=-1)
	{
		getftime(sema,&ftimep);
		close(sema);
		d_date.da_year=ftimep.ft_year+1980;
		d_date.da_mon=ftimep.ft_month;
		d_date.da_day=ftimep.ft_day;
		d_time.ti_hour=ftimep.ft_hour;
		d_time.ti_min=ftimep.ft_min;
		d_time.ti_sec=ftimep.ft_tsec;
		d_time.ti_hund=0;
		sematime=dostounix(&d_date,&d_time);		/* filetime naar seconden */
	}
	j=sprintf(pstr,"Synching on:");
	for (i=0;i<nrmatchakas;i++)
	{
		j=j+sprintf(&pstr[j]," %u:%u/%u",
					matchaka[i].zone,matchaka[i].net,matchaka[i].node);
		if (matchaka[i].point) j=j+sprintf(&pstr[j],".%u",matchaka[i].point);
		if (j>sizeof(pstr)-10) break;
	}
	sprintf(&pstr[j],"\n");
	pstring(pstr,0);

	connectpos=buflen-1;

	do
	{
		if (!seek_connect_back(&connectpos)) return(0);
		connectsfound++;
		ring=FALSE;
		datepos=emsipos=connectpos;
		if (!seekdate_back(&datepos,&ring)) return(0);
		if (!seek_emsi(&emsipos,buflen,ring)) return(0);
		gethisakas(emsipos);
		i=0;
		i=i+sprintf(pstr,"Found %u:%u/%u",
							hisakas[0].zone,hisakas[0].net,hisakas[0].node);
		if (hisakas[0].point) i=i+sprintf(&pstr[i],".%u",hisakas[0].point);
		if (nrhisakas>1) i=i+sprintf(&pstr[i]," Aka:");
		for (j=1;j<nrhisakas;j++)
		{
			i=i+sprintf(&pstr[i]," %u:%u/%u",
							hisakas[j].zone,hisakas[j].net,hisakas[j].node);
			if (hisakas[j].point) i=i+sprintf(&pstr[i],".%u",hisakas[j].point);
			if (i>sizeof(pstr)-10) break;
		}
		sprintf(&pstr[i],"\n");
		pstring(pstr,0);

		tzutcpos=tranxpos=emsipos;
		trx=seek_tranx(&tranxpos,buflen);
		if (trx==0)
		{
			sprintf(pstr,"No TRX# found in EMSI handshake.\n");
			pstring(pstr,0);
			continue;
		}
		tzutc=seek_tzutc(&tzutcpos,buflen);

		if (tzutc==MAXINT)
		{
			sprintf(pstr,"His TRX#: %lX. His time zone: UNKNOWN\n",trx);
			tzutc=0;
		}
		else sprintf(pstr,"His TRX#: %lX. His time zone: %ld (seconds)\n",trx,tzutc);
		pstring(pstr,0);

		t=get_unix_time(datepos,emsipos);
		mytzutc=getenv("TZUTC");
		if (mytzutc!=NULL)
		{
			sscanf(mytzutc,"%d",&i);
			j=abs(i)%100;
			i=i/100;
			mytz=60*(j+(60*i));
		}
		timediff=trx-tzutc-(t-mytz);
		sprintf(pstr,"Time difference in seconds: %ld\n",timediff);
		pstring(pstr,0);
		if (timediff>0) td=timediff; else td=0-timediff;
		if (t<sematime)
		{
			sprintf(pstr,"Connect found predates last correction by %s.\n",PROGNAME);
			pstring(pstr,0);
			break;				/* Should not set it twice on same info */
		}
		timer=time(NULL);
		if (timer-t>Maxage)
		{
			sprintf(pstr,"Info too old (%ld secs, %ld is max.).\n",timer-t,Maxage);
			pstring(pstr,0);
			break;				/* info too old */
		}
		if ((i=akasmatch())>=0 && td<=Maxdiff)
		{
			timer=time(NULL);
			timer=timer+timediff;
			unixtodos(timer,&d_date,&d_time);
			settime(&d_time);
			setdate(&d_date);
			j=sprintf(pstr,"Dos clock synchronised to %u:%u/%u",
				hisakas[i].zone,hisakas[i].net,hisakas[i].node);
			if (hisakas[i].point) sprintf(&pstr[j],".%u",hisakas[i].point);
			strcat(pstr,"\n");
			pstring(pstr,0);
			if (Rtcflag)
			{
				regs.h.ah=0x2c;	regs.h.al=0;
				intdos(&regs,&regs);	/* get system time */
				regs.h.ah=0x2d;	regs.h.al=0;
				intdos(&regs,&regs);	/* write back, also to CMOS clock */
				regs.h.ah=0x2a;	regs.h.al=0;
				intdos(&regs,&regs);	/* get system date */
				regs.h.ah=0x2b; 	regs.h.al=0;
				intdos(&regs,&regs);	/* write back, also to CMOS clock */
				sprintf(pstr,"Real Time Clock set to Dos clock.\n");
				pstring(pstr,0);
			}
			fsema=fopen(Semaphore,"w");	/* create or touch the semaphore */
			if (fsema!=NULL)
			{
				if (timediff<0) timer=timer-timediff;
				unixtodos(timer+2,&d_date,&d_time);
				ftimep.ft_year=d_date.da_year-1980;
				ftimep.ft_month=d_date.da_mon;
				ftimep.ft_day=d_date.da_day;
				ftimep.ft_hour=d_time.ti_hour;
				ftimep.ft_min=d_time.ti_min;
				ftimep.ft_tsec=d_time.ti_sec;
				setftime(fileno(fsema),&ftimep);
				fclose(fsema);
			}
			return(1);
		}
		connectpos-=2;
	}	while (connectsfound<Maxconnects && connectpos>0);

	if (envtz==NULL) putenv("TZ="); else putenv(envtzsave);
	if (logisopen) fclose(Outlog);
	return(0);
}

boolean seek_connect_back(int *pos)
{
	int p;

	p=*pos;

	while (p>0)
	{
		while(logbuf[p]!='\n' && p>0) p--;
		if (strncmp(&logbuf[p+13],"CONNECT",7)==0)
		{
			*pos=++p;
			return(TRUE);
		}
		else p--;
	}
	return(FALSE);
}

boolean seekdate_back(int *pos, boolean *ring)
{
	int p;
	boolean stoplooking_for_ring=FALSE;

	p=*pos;
	while (p>0)
	{
		while(logbuf[p]!='\n' && p>0) p--;
		if (strncmp(&logbuf[p+1],"----------",10)==0)
		{
			*pos=++p;
			return(TRUE);
		}
		else
		{
			if (!stoplooking_for_ring)
			{
				if (logbuf[p+1]!='=') stoplooking_for_ring=TRUE;
				else if (strncmp(&logbuf[p+13],"RING",4)==0) *ring=TRUE;
			}
			p--;
		}
	}
	return(FALSE);
}


boolean seek_emsi(int *pos, int buflen, boolean ring)
{
	int p;

	p=*pos;
	while (p<buflen)
	{
		while (logbuf[p]!='\n' && p<buflen)  p++;
		p++;
		if (p>=buflen) return(FALSE);
		if (strncmp(&logbuf[p+12],"{EMSI}",6)==0)
		{
			if (ring) { *pos=p; return(TRUE); }
			else if (!seek_emsi(&p,buflen,TRUE)) return(FALSE);
			*pos=p;
			return(TRUE);
		}
	}
	return(FALSE);
}

long int seek_tranx(int *trxpos,int buflen)
{
	int p,i=0, j;
	long trx=0;

	p=*trxpos;

	while (logbuf[p]=='~' && p<buflen)
	{
		i=0;
		while (logbuf[p+i]!='\n' && p+i<buflen)
		{
			if (strncmp(&logbuf[p+i],"{TRX#}",6)==0)
			{
				j=p+i+6;
				while (logbuf[j]!='{' && j<buflen)
				{
					if (logbuf[j]=='\n') j+=12;
					j++;
				}
				j++;
				if (logbuf[j]!='[') return(0);
				j++;
				sscanf(&logbuf[j],"%X",&trx);
				break;
			}
			else i++;
		}
		p=p+i+1;
		if (trx!=0) break;
	}
	*trxpos=p;
	return(trx);
}

long int seek_tzutc(int *tzpos,int buflen)
{
	int p,i=0, j, tz=MAXINT, hour, minute;

	p=*tzpos;

	while (logbuf[p]=='~' && p<buflen)
	{
		i=0;
		while (logbuf[p+i]!='\n' && p+i<buflen)
		{
			if (strncmp(&logbuf[p+i],"{TZUTC}",7)==0)
			{
				j=p+i+7;
				while (logbuf[j]!='{' && j<buflen)
				{
					if (logbuf[j]=='\n') j+=12;
					j++;
				}
				j++;
				if (logbuf[j]!='[') return(0);
				j++;
				sscanf(&logbuf[j],"%d",&tz);
				break;
			}
			else i++;
		}
		p=p+i+1;
		if (tz!=MAXINT) break;
	}
	*tzpos=p;
	if (tz!=MAXINT)
	{
		hour=tz/100;
		minute=abs(tz)%100;
		tz=60*(minute+(60*hour));
	}
	return(tz);
}

long int get_unix_time(int datepos, int emsipos)
{
	int i,year,day,hour,minute,second;
	char month[5];
	struct time d_time;
	struct date d_date;
	time_t t=0;

	sscanf(&logbuf[emsipos+2],"%d:%d:%d",&hour,&minute,&second);
	sscanf(&logbuf[datepos+16],"%d %s %d",&day,&month,&year);
	d_date.da_day=day;
	d_time.ti_hour=hour; d_time.ti_min=minute; d_time.ti_sec=second;
	for (i=0;i<=12;i++) if (stricmp(month,monthn[i])==0) break;

	if (year<70) d_date.da_year=year+2000;
	else if (year<100) d_date.da_year=year+1900; else d_date.da_year=year;
	d_date.da_mon=i;
	d_time.ti_hund=0;
	t=dostounix(&d_date,&d_time);
	return(t);
}

int akasmatch(void)
{
	int i,j,r=-1;

	for (i=0;i<nrmatchakas;i++)
	{
		for (j=0;j<nrhisakas;j++)
		{
			if (matchaka[i].zone==hisakas[j].zone &&
				matchaka[i].net==hisakas[j].net &&
				matchaka[i].node==hisakas[j].node &&
				matchaka[i].point==hisakas[j].point)
				{ r=j; break ; }
		}
	}
	return(r);
}

void gethisakas(int emsipos)
{
	int i,j;

	nrhisakas=0;
	i=emsipos+19;
	do
	{
		while (isspace(logbuf[i])) {if (logbuf[i]=='\n') i+=12; i++; }
		j=scanfidonumber(&logbuf[i],&hisakas[nrhisakas]);
		if (j>0)
		{
			nrhisakas++;
			i=j+i;
			if (logbuf[i]=='\n') i+=12;
		}
	}	while (nrhisakas<NRHISAKAS && j>0 && isspace(logbuf[i]));
}

int getcommands(int nargs, char *args[])
{
	int i,j,k,c;

	for (i=1;i<nargs;i++)
	{
		c=toupper(args[i][0]);
		switch (c)
		{
			case 'A' :	if (args[i][1]=='=')
							{
								sscanf(&args[i][2],"%ld",&Maxage);
							}
							break;
			case 'C' :	if (args[i][1]=='=')
							{
								sscanf(&args[i][2],"%d",&Maxconnects);
							}
							break;
			case 'D' :	if (args[i][1]=='=')
							{
								sscanf(&args[i][2],"%ld",&Maxdiff);
							}
							break;
			case 'I'	:	if (args[i][1]=='=')
							{
								strncpy(Inputlog,&args[i][2],sizeof(Inputlog)-1);
							}
							break;
			case 'L'	:	if (args[i][1]=='=')
							{
								strncpy(Outputlog,&args[i][2],sizeof(Outputlog)-1);
							}
							break;
			case 'N'	:	if (args[i][1]=='=') j=scanfidonumber(&args[i][2],&Mynode);
							break;
			case 'R'	:	Rtcflag=TRUE;
							break;
			case 'S'	:	if (args[i][1]=='=' && nrmatchakas<NRMATCHAKAS)
							{
								k=0;
								do
								{
								j=scanfidonumber(&args[i][2+k],&matchaka[nrmatchakas]);
								nrmatchakas++;
								k=k+j+1;
								}	while (args[i][k+1]==',' && nrmatchakas<NRMATCHAKAS);
							}
							break;
			default	:	help(); return(0);

		}
	}

	return(1);
}

void logheader(void)
{
	struct tm *local;
	time_t l;
	int year;

	if (logisopen || logflag<=0) return;

	if ((Outlog=fopen(Outputlog,"at"))!=NULL)
	{
		time(&l);
		local = localtime(&l);
		year=local->tm_year;
		if (year>=100) year+=1900;

		fprintf(Outlog,"\n----------  %s %02d %s %d, %s v %s\n",
		dow[local->tm_wday], local->tm_mday, monthn[local->tm_mon + 1],
																year,PROGNAME,VERSION);
		logisopen=TRUE;
	}
	else logflag=0;
}


/* timestamp to log file */

void timestamp(FILE *f, int s)
{
	const char ind[]={'#','!','?','$','*',':','+','-','%','~','='};
	time_t l;
	struct tm *loc;

	time(&l);
	loc=localtime(&l);
	fprintf(f,"%c %2d:%02d:%02d  ",ind[s],loc->tm_hour,loc->tm_min,loc->tm_sec);
}

/* string output to screen and to log */

void pstring(char *st,int n)
{
	printf("%s",st);
	if (logflag>0)
	{
		if (n>=0) timestamp(Outlog,n);
		fprintf(Outlog,"%s",st);
	}
}

int scanfidonumber(char *s, nodenumber *f)
{
	char fn[20];
	char nr[10];
	char s1,s2;
	int i,j,l;
	unsigned int n;
	boolean done=FALSE;

	f->zone=Lastnode.zone;
	f->net=Lastnode.net;
	f->node=Lastnode.node;
	f->point=0;

	i=0;
	while (s[i] && !isspace(s[i]) && s[i]!='}' && s[i]!=',' &&
							  s[i]!='@' && i<sizeof(fn)-1) { fn[i]=s[i]; i++; }
	fn[i]=0;
	l=i;
	if (s[l]=='@')			/* skip over domain name if present */
	{
		while (s[l] && !isspace(s[l]) && s[l]!='}' && s[l]!=',') l++;
	}

	i=0;
	s2=s1=0;
	if (fn[0]=='.')
	{
		if (fn[1]==0) { f->point=Lastnode.point=Mynode.point; return(l); }
		if (!isdigit(fn[1])) return(-1);
		i++;
		s2='.';
	}

	while (!done)
	{
		s1=s2;
		j=0;
		while (isdigit(fn[i]) && j<sizeof(nr)-1) {nr[j]=fn[i]; i++; j++; }
		nr[j]=0;
		s2=fn[i];
		if (sscanf(nr,"%u",&n)!=1) return(-1);
		i++;
		done=(s2==0);

		switch (s1)
		{
			case 0	:	switch (s2)
							{
								case 0  :	f->node=n; break;
								case '.':	f->node=n; break;
								case '/':	f->net=n; break;
								case ':':	f->zone=n; break;
								default :	return(-1);
							}
							break;
			case '.'	:	if (s2!=0) return(-1);
							else f->point=n;
							break;
			case '/'	:	if (s2==0 || s2=='.') f->node=n;
							else return(-1);
							break;
			case ':'	:	if (s2=='/') f->net=n;
							else return(-1);
							break;
			default	:	return(-1);
		}
	}
	Lastnode.zone=f->zone;
	Lastnode.net=f->net;
	Lastnode.node=f->node;
	Lastnode.point=f->point;
	return(l);
}


void help(void)
{
	printf("Usage   : %s [command=parameter][..]\n",PROGNAME);
	printf("Commands:\n\n");
	printf("   a=<maxage>        Maximum age of connect (secs). {%d}\n",MAXAGE);
	printf("   c=<maxconnects>   Max number connects to search.    {%d}\n",MAXCONNECT);
	printf("   d=<maxdiff>       Max of difference (secs).       {%d}\n",MAXDIFF);
	printf("   i=<filespec>      Name of mailer log file.     {IM.LOG}\n");
	printf("   l=<filespec>      Name of Mtick log file.  {MTSYNC.LOG}\n");
	printf("   n=<nodenr>        Your main node number.\n");
	printf("   s=<nodenr>[,nr].. Nodenumber(s) to sync on.\n");
	printf("   r                 Update real time clock also.\n");

}
